#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include "grab.h"
#include "colorspace.h"
#include "sound.h"
#include "writeavi.h"
#include "capture.h"

/*-------------------------------------------------------------------------*/

void
fifo_init(struct FIFO *fifo, char *name, int slots)
{
    pthread_mutex_init(&fifo->lock, NULL);
    pthread_cond_init(&fifo->hasdata, NULL);
    fifo->name  = name;
    fifo->slots = slots;
    fifo->read  = 0;
    fifo->write = 0;
    fifo->eof   = 0;
}

int
fifo_put(struct FIFO *fifo, unsigned char *data, int size)
{
    pthread_mutex_lock(&fifo->lock);
    if (debug > 1)
	fprintf(stderr,"put %s %p %d\n",fifo->name,data,size);
    if (NULL == data) {
	fifo->eof = 1;
	pthread_cond_signal(&fifo->hasdata);
	pthread_mutex_unlock(&fifo->lock);
	return 0;
    }
    if ((fifo->write + 1) % fifo->slots == fifo->read) {
	pthread_mutex_unlock(&fifo->lock);
	fprintf(stderr,"fifo %s is full\n",fifo->name);
	return 0;
    }
    fifo->data[fifo->write] = data;
    fifo->size[fifo->write] = size;
    fifo->write++;
    if (fifo->write >= fifo->slots)
	fifo->write = 0;
    pthread_cond_signal(&fifo->hasdata);
    pthread_mutex_unlock(&fifo->lock);
    return size;
}

int
fifo_get(struct FIFO *fifo, unsigned char **data)
{
    int size;

    pthread_mutex_lock(&fifo->lock);
    if (debug > 1)
	fprintf(stderr,"get %s\n",fifo->name);
    while (fifo->write == fifo->read && 0 == fifo->eof) {
	pthread_cond_wait(&fifo->hasdata, &fifo->lock);
    }
    if (fifo->write == fifo->read) {
	*data = NULL;
	pthread_cond_signal(&fifo->hasdata);
	pthread_mutex_unlock(&fifo->lock);
	return 0;
    }
    *data = fifo->data[fifo->read];
    size = fifo->size[fifo->read];
    fifo->read++;
    if (fifo->read >= fifo->slots)
	fifo->read = 0;
    pthread_cond_signal(&fifo->hasdata);
    pthread_mutex_unlock(&fifo->lock);
    return size;
}

static void*
flushit(void *arg)
{
    int old;

    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,&old);
    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,&old);
    for (;;) {
	sleep(1);
	sync();
    }
    return NULL;
}

/*-------------------------------------------------------------------------*/

static pthread_mutex_t  avi_lock;
struct MOVIE_PARAMS     avi_par;
static pthread_t        taudio;
static pthread_t        tvideo;
static pthread_t        tflush;
static struct FIFO      faudio;
static struct FIFO      fvideo;
static char             *baudio,*bvideo;
static int              saudio,svideo;
static int              iaudio,ivideo;
static int              naudio,nvideo;
static int              caudio,cvideo;

static void*
avi_audio_writer(void *arg)
{
    struct FIFO *fifo = arg;
    unsigned char *data;
    int size;

    if (debug)
	fprintf(stderr,"avi_audio_writer start\n");
    for (;;) {
	size = fifo_get(fifo,&data);
	if (NULL == data)
	    break;
	pthread_mutex_lock(&avi_lock);
	avi_writesound(data, size);
	pthread_mutex_unlock(&avi_lock);
    }
    if (debug)
	fprintf(stderr,"avi_audio_writer done\n");
    return NULL;
}

static void *
avi_video_writer(void *arg)
{
    struct FIFO *fifo = arg;
    unsigned char *data;
    int size;

    if (debug)
	fprintf(stderr,"avi_video_writer start\n");
    for (;;) {
	size = fifo_get(fifo,&data);
	if (NULL == data)
	    break;
	pthread_mutex_lock(&avi_lock);
	avi_writeframe(data, size);
	pthread_mutex_unlock(&avi_lock);
    }
    if (debug)
	fprintf(stderr,"avi_video_writer done\n");
    return NULL;
}

int
avi_init_writer(char *filename, struct MOVIE_PARAMS *params,
		int slots, int *sound)
{
    int linelength;

    if (debug)
	fprintf(stderr,"avi_init_writer\n");
    pthread_mutex_init(&avi_lock, NULL);

    /* audio */
    *sound = -1;
    if (params->channels > 0)
	*sound = sound_open(params);
    if (params->channels > 0) {
	fifo_init(&faudio,"audio",slots);
	pthread_create(&taudio,NULL,avi_audio_writer,&faudio);
	iaudio = 0;
	naudio = slots+2;
	saudio = sound_bufsize();
	baudio = malloc(saudio*naudio);
	caudio = 0;
    }

    /* video */
    grabber_setparams(params->video_format, &params->width, &params->height,
		      &linelength,0,0);
    fifo_init(&fvideo,"video",slots);
    pthread_create(&tvideo,NULL,avi_video_writer,&fvideo);
    ivideo = 0;
    nvideo = slots+2;
    svideo = params->width * params->height *
	format2depth[params->video_format]/8;
    if (0 == svideo)
	svideo = params->width * params->height * 3;
    bvideo = malloc(svideo*nvideo);
    cvideo = 0;
    
    /* init */
    memcpy(&avi_par,params,sizeof(struct MOVIE_PARAMS));
    avi_open(filename,params);
    
    return 0;
}

int
avi_start_writer()
{
    if (debug)
	fprintf(stderr,"avi_start_writer\n");
    if (grabber->grab_start)
	grabber->grab_start(avi_par.fps,0);
    if (avi_par.channels > 0)
	sound_startrec();
    pthread_create(&tflush,NULL,flushit,NULL);
    return 0;
}

int
avi_stop_writer()
{
    void *dummy;
    long long ausec,vusec;
    int soundbytes;

    if (debug)
	fprintf(stderr,"avi_stop_writer\n");

    if (avi_par.channels > 0) {
	vusec = cvideo * 1000 / avi_par.fps;
	ausec = ((long long)caudio*saudio*8*1000) /
	    (avi_par.rate*avi_par.bits*avi_par.channels);
	while (vusec < ausec) {
	    grab_put_video();
	    vusec = cvideo * 1000 / avi_par.fps;
	}
	soundbytes = (int)(vusec-ausec)*
	    avi_par.rate*avi_par.bits*avi_par.channels/8/1000;
	soundbytes = (soundbytes+4) & ~0x03;
	fprintf(stderr,"vs=%Ld as=%Ld %d/%d\n",vusec,ausec,
		soundbytes,saudio);
	while (soundbytes > saudio) {
	    grab_put_audio();
	    soundbytes -= saudio;
	}
	sound_read(baudio + iaudio*saudio);
	fifo_put(&faudio,baudio + iaudio*saudio,soundbytes);
    }

    /* send EOF + join threads */
    if (avi_par.channels > 0) {
	fifo_put(&faudio,NULL,0);
	pthread_join(taudio,&dummy);
    }
    fifo_put(&fvideo,NULL,0);
    pthread_join(tvideo,&dummy);
    pthread_cancel(tflush);
    pthread_join(tflush,&dummy);

    /* close file */
    avi_close();
    if (grabber->grab_stop)
	grabber->grab_stop();
    if (avi_par.channels > 0)
	sound_close();
    return 0;
}

/*-------------------------------------------------------------------------*/

int
generic_init_writer(void* (*video_writer)(void *arg),
		    int format, int *width, int *height, int slots)
{
    int linelength;

    if (debug)
	fprintf(stderr,"generic_init_writer\n");
    
    grabber_setparams(format, width, height, &linelength, 0, 0);
    fifo_init(&fvideo,"video",slots);
    pthread_create(&tvideo,NULL,video_writer,&fvideo);
    ivideo = 0;
    nvideo = slots+2;
    svideo = *width * *height * format2depth[format]/8;
    if (0 == svideo)
	svideo = *width * *height * 3;
    bvideo = malloc(svideo*nvideo);
    cvideo = 0;
    
    return 0;
}

int
generic_start_writer(int fps)
{
    if (debug)
	fprintf(stderr,"generic_start_writer\n");
    if (grabber->grab_start)
	grabber->grab_start(fps,0);
    pthread_create(&tflush,NULL,flushit,NULL);
    return 0;
}

int
generic_stop_writer()
{
    void *dummy;

    if (debug)
	fprintf(stderr,"generic_stop_writer\n");
    fifo_put(&fvideo,NULL,0);
    pthread_join(tvideo,&dummy);
    pthread_cancel(tflush);
    pthread_join(tflush,&dummy);
    if (grabber->grab_stop)
	grabber->grab_stop();
    return 0;
}

/*-------------------------------------------------------------------------*/

int
grab_put_video()
{
    int size;

    if (debug)
	fprintf(stderr,"grab_put_video\n");

    /* get next frame */
    grabber_capture(bvideo + ivideo*svideo,0,&size);

    fifo_put(&fvideo,bvideo + ivideo*svideo,size);
    ivideo = (ivideo+1) % nvideo;
    cvideo++;
    return 0;
}

int
grab_put_audio()
{
    if (debug)
	fprintf(stderr,"grab_put_audio\n");

    sound_read(baudio + iaudio*saudio);
    fifo_put(&faudio,baudio + iaudio*saudio,saudio);
    iaudio = (iaudio+1) % naudio;
    caudio++;
    return 0;
}
